#define _GNU_SOURCE

#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/sysinfo.h>
#include <sysexits.h>
#include <threads.h>
#include <unistd.h>

#define EXAMPLE_PORT	29292
#define MAX_EVENTS		8

static int worker(void* _data)
{
	(void)_data;
	int efd;
	int sk;
	int res;
	struct sockaddr_in addr;
	struct epoll_event event;
	struct epoll_event events[MAX_EVENTS];
	int optval;

    efd = epoll_create1(0);
	if (efd == -1)
	{
		perror("epoll_create1");
		exit(EX_OSERR);
	}

	sk = socket(PF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);
    if (sk == -1)
	{
		perror("socket");
		exit(EX_OSERR);
	}
	optval = 1;
	res = setsockopt(sk, SOL_SOCKET, SO_REUSEPORT, &optval, sizeof optval);
	if (res == -1)
	{
		perror("setsockopt");
		exit(EX_OSERR);
	}

	memset(&addr, 0, sizeof addr);
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(EXAMPLE_PORT);
    res = bind(sk, (struct sockaddr*)&addr, sizeof addr);
	if (res == -1)
	{
		perror("bind");
		exit(EX_OSERR);
	}

	res = listen(sk, 0);
	if (res == -1)
	{
		perror("listen");
		exit(EX_OSERR);
	}

	event.events = EPOLLIN;
	event.data.fd = sk;
	res = epoll_ctl(efd, EPOLL_CTL_ADD, sk, &event);
	if (res == -1)
	{
		perror("epoll_ctl");
		exit(EX_OSERR);
	}

	for (;;)
	{
		res = epoll_wait(efd, events, MAX_EVENTS, -1);
		if (res == -1)
		{
			perror("epoll_wait");
			exit(EX_OSERR);
		}

		for (int i = 0; i < res; i++)
		{
			if (events[i].data.fd == sk)
			{
				res = accept4(sk, NULL, NULL, SOCK_NONBLOCK);
				event.events = EPOLLIN;
				event.data.fd = res;
				res = epoll_ctl(efd, EPOLL_CTL_ADD, res, &event);
				if (res == -1)
				{
					perror("epoll_ctl");
					exit(EX_OSERR);
				}
				printf("Thread %lx accepted new client with socket %d\n", thrd_current(), event.data.fd);
			} else
			{
				close(events[i].data.fd);
				printf("Thread %lx closed client socket %d\n", thrd_current(), event.data.fd);
			}
		}
	}

    close(sk);
	close(efd);
}

int main(int _argc, char** _argv)
{
	(void)_argc;
	(void)_argv;
	int cpus;
	thrd_t* threads;

    cpus = get_nprocs();
	threads = malloc(cpus * sizeof *threads);
	for (int i = 0; i < cpus; i++)
		thrd_create(&threads[i], worker, NULL);

	for (int i = 0; i < cpus; i++)
		thrd_join(threads[i], NULL);

	exit(EX_OK);
}
